-module(tcp).
-export([start/0, stop/0]).
-export([listen/0]).

%% gen_server callbacks.
-export([init/1,
         handle_call/3, handle_cast/2, handle_info/2,
         terminate/2, code_change/3]).

-define('DRIVER_NAME', 'tcp_drv').

-define(DRV_LISTEN, 1).

-define(debug, 1).
%% Debug log.
-ifdef(debug).
-define(log(Msg, MsgArgs),
    io:format("[~p:~p] ~s\n", [?MODULE, ?LINE, io_lib:format(Msg, MsgArgs)])).
-else.
-define(log(Msg, MsgArgs), true).
-endif.

-behaviour(gen_server).

-record(state, {port}).

start() ->
    gen_server:start({local, ?MODULE}, ?MODULE, [], []).

stop() ->
    complex ! stop.

listen() ->
    gen_server:call(?MODULE, {listen}).

%%%===================================================================
%%% gen_server callbacks
%%%===================================================================

%%--------------------------------------------------------------------
%% @private
%% @doc Handle start.
%%
%% @spec init(Args) ->
%%          {ok, State} |
%%          {ok, State, Timeout} |
%%          {ok, State, hibernate} |
%%          {stop, Reason} |
%%          ignore
%% @end
%%--------------------------------------------------------------------
init([]) ->
    process_flag(trap_exit, true),
    DirName = re:replace(filename:dirname(code:which(?MODULE)), 
                         "/?[^/]+/\\.\\.", "", [{return,list}]),
    SearchDir = filename:join(filename:dirname(DirName), "ebin"),
    ?log("init, lib path: ~s", [SearchDir]),
    try erl_ddll:load(SearchDir, ?DRIVER_NAME) of
    ok ->
        Port = open_port({spawn_driver, ?DRIVER_NAME}, [binary]),
        {ok, #state{port=Port}};
    {error, Reason} ->
        throw(erl_ddll:format_error(Reason))
    catch _:Error ->
        {stop, Error}
    end.

%%--------------------------------------------------------------------
%% @private
%% @doc Handle synchronous call.
%%
%% @spec handle_call(Request, From, State) ->
%%          {reply, Reply, NewState} |
%%          {reply, Reply, NewState, Timeout} |
%%          {reply, Reply, NewState, hibernate} |
%%          {noreply, NewState} |
%%          {noreply, NewState, Timeout} |
%%          {noreply, NewState, hibernate} |
%%          {stop, Reason, Reply, NewState} |
%%          {stop, Reason, NewState}
%% @end
%%-------------------------------------------------------------------

% No need to support context termination - context allocation
% is handled on port creation, and the resource will be 
% automatically reclaimed upon death of the port driver.  
%handle_call({term}, _From, State) ->
%    ?log("~p", [term]),
%    Message = <<(?ZMQ_TERM):8>>,
%    Reply = driver(State#state.port, Message),
%    {reply, Reply, State};

handle_call({listen}, _From, State) ->
    ?log("handle_call ~p", [listen]),
    do_call(State, encode_listen());

handle_call(port, _From, #state{port = Port} = State) ->
    {reply, Port, State};

handle_call(Request, _From, State) ->
    {stop, {unknown_call, Request}, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc Handle asynchronous call.
%%
%% @spec handle_cast(Msg, State) ->
%%          {noreply, NewState} |
%%          {noreply, NewState, Timeout} |
%%          {noreply, NewState, hibernate} |
%%          {stop, Reason, NewState}
%% @end
%%--------------------------------------------------------------------
handle_cast(Msg, State) ->
    {stop, {unknown_cast, Msg}, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc Handle message.
%%
%% @spec handle_info(Info, State) ->
%%          {noreply, NewState} |
%%          {noreply, NewState, Timeout} |
%%          {noreply, NewState, hibernate} |
%%          {stop, Reason, NewState}
%% @end
%%--------------------------------------------------------------------
handle_info(_Info, State) ->
    ?log("unhandled message: ~p\n", [_Info]),
    {noreply, State}.

%%--------------------------------------------------------------------
%% @private
%% @doc Handle termination/shutdown.
%%
%% @spec terminate(Reason, State) -> void()
%% @end
%%--------------------------------------------------------------------
terminate(_Reason, State) ->
    port_close(State#state.port),
    ok.

%%--------------------------------------------------------------------
%% @private
%% @doc Handle code change.
%%
%% @spec code_change(OldVsn, State, Extra) -> 
%%          {ok, NewState}
%% @end
%%--------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%%===================================================================
%%% Internal functions
%%%===================================================================

encode_listen() ->
    <<(?DRV_LISTEN):8>>.

do_call(#state{} = State, Message) ->
    Reply = driver(State#state.port, Message),
    {reply, Reply, State}.

driver(Port, Message) ->
    ?log("port command ~p", [Message]),
    port_command(Port, Message),
    receive
        ok ->
            ok;
        {ok, Term} ->
            {ok, Term};
        Err = {error, _} ->
            Err
    end.
